#ifndef TRXPARTICLE_H
#define TRXPARTICLE_H
/** \file
   Definitions for particles.
*/
/*
   
   Copyright  1992-2015 M.Kraemer, GSI Darmstadt 

   $Log: trxparticle.h,v $
   Revision 1.6  2015/01/05 15:36:48  kraemer
   streamline 64bit, complete rework for mixtures, nuclear, Auger's

   Revision 1.5  2009/05/29 12:40:44  kraemer
   reworked projectiles, targets

   Revision 1.4  2008/06/20 14:34:24  kraemer
   tilted vol's

   Revision 1.3  2008/02/05 10:55:33  kraemer
   doxygenized

   Revision 1.2  2002/11/25 16:31:04  kraemer
   record start loc,dir; particleevt w/ header; kinematics macro

   Revision 1.1  2001/01/09 14:54:35  kraemer
   *** empty log message ***


*/

#include "cmderr.h"
#include "trxrnd.h"
#include "trxxs.h"
#include "trxvol.h"
#include "trxgeo.h"
#include "trxsrc.h"
                                      /* common masks and flags */
                                      /* masks */
#define TRXPARTICLE_MASKGLOBAL 0xFF000000  /* global (e.g. debug) */
#define TRXPARTICLE_MASKFUNC   0x00FF0000  /* function specific flags */
#define TRXPARTICLE_MASKQUAL   0x0000FFFF  /* for additional qualifiers */

                                      /* flags */
#define TRXPARTICLE_DEBUG      0x80000000  /* general check flag */
#define TRXPARTICLE_DEBUG2     0x40000000  /* general check flag */
#define TRXPARTICLE_CHECK      TRXPARTICLE_DEBUG   /* general check flag */
#define TRXPARTICLE_QUIET      0x10000000  /* suppress error messages */
#define TRXPARTICLE_NOMSG      TRXPARTICLE_QUIET

                                      /** \struct STRXPARTICLE
                                          \brief Describes a particle.
                                      */ 
struct STRXPARTICLE {
   unsigned int uiFlag;               /**< flag bits */    
   unsigned int uiEvt;                /**< primary event number */    
   unsigned int uiID;                 /**< unique particle ID */
   unsigned int uiIASequence;         /**< interaction sequence number */ 
   int iIA;                           /**< current interaction type */
   struct STRXXSPROJ *psXSP;          /**< projectile cross section for this projectile (ref., don't free()) */
   struct STRXXS *psXS;               /**< actual cross section to be used (ref., don't free()) */
   int iSh;                           /**< target shell affected (==0 if unspecified/all shells) */
   unsigned int uiVolSrc;             /**< source volume of primary particle */
   ssize_t lvVoxel[3];                /**< current voxel indices (1..n), in case we're running through a voxel grid */
   struct STRXVOL *psVol;             /**< current volume (ref., don't free()) */
   struct STRXVOL *psVolExcl;         /**< volume excluded from isect (ref., don't free()) */
   unsigned int uiSrf;                /**< current surface number, 1..n */
   unsigned int uiSrfExcl;            /**< surface excluded from isect */
   double dSigTot;                    /**< total x-sect for next interaction */
   double dMFP, dIAct;                /**< mean,actual free path */
   double dDist;                      /**< distance to intersection */
   double dvx[4], dvx0[4];            /**< x,t actual, original */
   double dvd[4], dvd0[4];            /**< direction vector actual, original */                     
   double dE, dE0;                    /**< kinetic energy actual, original
                                           [keV] (x,e-,e+)
                                           [MeV/u] (ions)       
                                      */
   double dbeta2, dbeta, dgamma, dT;  /**< (v/c)**2, (v/c), Lorentz factor, kinetic energy [keV] */
   double ddE;                        /**< energy loss for current interaction [keV] */
   struct STRXPROJ sP;                /**< projectile */
};

                                      /* common macro to derive kinematics from kinetic energy */
#define TRXPARTICLEKINEMATIC(p)\
   switch((p).sP.uiType ) {\
   case TRXPROJ_ION:\
      (p).dgamma = (p).dT / ( TRX_AMU * 1.0E3 * (p).sP.sZA.TRXZA_A ) + 1.0E0;\
      (p).dbeta2 = 1.0E0 - ( 1.0E0 / (p).dgamma ) * ( 1.0E0 / (p).dgamma );\
      (p).dbeta  = sqrt((p).dbeta2);\
      (p).dE = (p).dT * 1E-3 / (p).sP.sZA.TRXZA_A;\
   break;\
   case TRXPROJ_ELECTRON:\
   case TRXPROJ_POSITRON:\
      (p).dgamma = (p).dT / TRX_M0 + 1.0E0;\
      (p).dbeta2 = 1.0E0 - ( 1.0E0 / (p).dgamma ) * ( 1.0E0 / (p).dgamma );\
      (p).dbeta  = sqrt((p).dbeta2);\
      (p).dE = (p).dT;\
   break;\
   case TRXPROJ_X:\
      (p).dgamma = DBL_MAX;\
      (p).dbeta2 = 1.0E0;\
      (p).dbeta  = sqrt((p).dbeta2);\
      (p).dE = (p).dT;\
   break;\
   default:\
      sCE.iRC = ERRTRXPARTICLE_BADPROJ; goto gError;\
   break;\
   }
/*

   General particle handling
   -------------------------

   TRXParticleChangeDir  ( partto, partfr, costh, phi, flag )
                         change particle direction vector
                       7 x to test
                         o calls:

   TRXParticleFree       ( part, flag ) 
                         free resources
                       4 x to test
                         o calls:
			                      o TRXProjFree()
			                      o TRXParticleInit()  
                         
   TRXParticleInit       ( part, flag ) 
                         initialize
                       3 x to test
                         o calls:
			                        o TRXProjInit()
			                        o TRXSpaceInit()

   TRXParticleInspect    ( part, flag )
                         inspect
                       1 x to test
                         o calls:
			                        o TRXProjInspect()

   TRXParticleScore      ( particle, maskPT, maskIA, flag )
                         score interaction & eloss
                       5 x to test
                         o calls:  
   
   TRXParticleSpace      ( part, flag )
                         inspect
                       2 o to do
                         o calls:
			                        o TRXProjSpace()

   TRXParticle3Product   ( part1, part2, part3, flag )
                         spat-product to test co-planarity
                       6 x to test
                         o calls  
  

   TRXParticleError  : error handling

*/

struct SCMDERR TRXParticleChangeDir( struct STRXPARTICLE *, struct STRXPARTICLE *, double, double, unsigned int );
struct SCMDERR TRXParticleFree     ( struct STRXPARTICLE *, unsigned int );
struct SCMDERR TRXParticleInit     ( struct STRXPARTICLE *, unsigned int );
struct SCMDERR TRXParticleInspect  ( struct STRXPARTICLE *, unsigned int );
struct SCMDERR TRXParticleScore    ( struct STRXPARTICLE *, unsigned int, unsigned int, unsigned int );
struct SCMDERR TRXParticleSpace    ( struct STRXPARTICLE *, unsigned int );
double TRXParticle3Product         ( struct STRXPARTICLE *, struct STRXPARTICLE *, struct STRXPARTICLE *, unsigned int );

struct STRXPARTICLE _TRXParticleTest( unsigned int ); 

                                      /** \struct STRXPARTICLESTACK
                                          \brief A stack of particle.
                                      */ 
struct STRXPARTICLESTACK {
#define TRXPARTICLESTACK_INC 8
   int nP, mP;                        /**< actual/max # particles */
   struct STRXPARTICLE *psP;          /**< particles */
};

/*

   Particle stack handling
   -------------------------

   TRXParticleStackFree   ( ps, flag ) 
                          o free resources
                        3 o to do
                          o calls:

   TRXParticleStackInit   ( ps, flag ) 
                          o initialize
                        2 o to do
                          o calls:

   TRXParticleStackInspect( ps, flag )
                          o inspect
                        1 o to do
                          o calls:


*/

struct SCMDERR TRXParticleStackFree     ( struct STRXPARTICLESTACK *, unsigned int );
struct SCMDERR TRXParticleStackInit     ( struct STRXPARTICLESTACK *, unsigned int );
struct SCMDERR TRXParticleStackInspect  ( struct STRXPARTICLESTACK *, unsigned int );

struct STRXPARTICLESTACK _TRXParticleStackTest( unsigned int ); 

                                      /** \struct STRXPARTICLEVTOLD
                                          \brief particle event output, old format.
                                      */ 
struct STRXPARTICLEEVTOLD {
   unsigned int uiEvt;               /**< primary event number */
   unsigned int uiID;                /**< particle ID */
   unsigned int ui32IASequence;      /**< interaction sequence number */ 
   unsigned int uiPT;                /**< particle type */
            int  iIA;                /**< interaction */
   unsigned int uiVol;               /**< Vol ID */
   double dvx0[4];                   /**< 0..2: location, 3: time */
   float rvd [3];                    /**< direction */
   float rE, rdE;                    /**< energy, energy loss */
};


#define TRXPARTICLEEVT_FILEVERSION 20020830

                                      /** \struct STRXPARTICLEVT
                                          \brief particle event output.
                                      */ 
struct STRXPARTICLEEVT {
   uint32_t ui32Version, ui32Length;  /**< version, length of this structure */
   uint16_t ui16ByteSex;              /**< =1 for native endian, else byte swap */
   uint16_t ui16Mode;                 /**< always=0 for particle evt */
   uint32_t ui32Evt;                  /**< primary event number */
   uint32_t ui32ID;                   /**< particle ID */
   uint32_t ui32IASequence;           /**< interaction sequence number */ 
   uint32_t ui32PT;                   /**< particle type */
    int32_t  i32IA;                   /**< interaction */
   uint32_t ui32Vol, ui32VolSrc;      /**< Vol ID actual, original */
   double dvx[4], dvx0[4];            /**< 0..2: location, 3: time; actual, original */
   float  rvd[3], rvd0[3];            /**< direction actual, original */
   float rE, rE0;                     /**< energy actual, original */
   float rdE;                         /**< energy loss */
};

/*
   Particle events for output
   --------------------------
    
   TRXParticleEvt         ( partevt, part, flag )
                          x to test
		                          o calls

   TRXParticleEvtInspect  ( partevt, flag )
                          x to test
                          o calls:


*/
                                      /* function prototypes */
int TRXParticleEvtInspect( struct STRXPARTICLEEVT *, unsigned int );
int TRXParticleEvt       ( struct STRXPARTICLEEVT *, struct STRXPARTICLE *, unsigned int );

struct STRXPARTICLEEVT _TRXParticleEvtTest( unsigned int ); 

/*

   Functions for particles & volumes
   -------------------------------
   
   TRXParticleVolISect      ( particle, vol, flag )
                              Query particle intersection with a volume:
                              resets/sets voxel indices, volume and surface info,
                              distance
                          4 . to test
                            o returns:
                              o  < 0: error
                              o  > 0: intersection
                              o == 0: no intersection
                            . tilted volumes
                            o calls
                              o TRXParticleVolISectSrf()
   
   TRXParticleVolISectSrf   ( particle, vol, srf, flag )
                              Query particle intersection with a surface of a volume
                          3 . to test
                              o to complete
                                o torus
                                o angular limits
                                . tilted volumes
                            o returns:
                              o  < 0: error
                              o  > 0: intersection
                              o == 0: no intersection
                            o calls
   
   TRXParticleVolLocate     ( particle, vol, flag )
                              Locate a particle within a volume,
                              resets/sets vol/surface/voxel indices
                          2 . to test
                              o vollocate.cyl
                              o to complete
                                o torus
                                o angular limits
                                . tilted volumes
                            o returns:
                              o  < 0: error
                              o  > 0: inside
                              o == 0: outside
                            o calls
                              o TRXParticleVolLocateSrf()

   TRXParticleVolLocateSrf  ( particle, vol, srf, flag )
                              Locate particle on a volume's surface
                          1 . to test 
                              o to complete 
                                o torus
                                o angular limits
                                . tilted volumes
                            o returns:
                              o  < 0: error
                              o  > 0: on surface
                              o == 0: not on surface
                            o calls
                              o TRXParticleVolLocateSrf() 

*/

                                      /* function action flags */
                                      /* don't isect recursively */ 
#define TRXPARTICLEVOL_NORECURSIVEISECT 0x00010000
                                      /* skip transformation */ 
#define TRXPARTICLEVOL_SKIPTRAFO        0x00020000

struct SCMDERR TRXParticleVolLocateSrf( struct STRXPARTICLE *, struct STRXVOL *, unsigned int, unsigned int );
struct SCMDERR TRXParticleVolLocate   ( struct STRXPARTICLE *, struct STRXVOL *, unsigned int );
struct SCMDERR TRXParticleVolISectSrf ( struct STRXPARTICLE *, struct STRXVOL *, unsigned int, unsigned int );
struct SCMDERR TRXParticleVolISect    ( struct STRXPARTICLE *, struct STRXVOL *, unsigned int );

/*

   Functions involving the geometry base
   -------------------------------------
   
   TRXParticleGeoISect   ( particle, geo, flag ); 
                         . to test
                         o calls 
                           o TRXParticleVolISect()

   TRXParticleGeoLocate  ( particle, geo, flag ); 
                         o locate particle in volume of geometry
                         . to test
                         o to complete
                           o support nested / multiple locations? 
                         o calls 
                           o TRXParticleVolLocate()
                                   
*/

struct SCMDERR TRXParticleGeoLocate   ( struct STRXPARTICLE *, struct STRXGEO *, unsigned int );
struct SCMDERR TRXParticleGeoISect    ( struct STRXPARTICLE *, struct STRXGEO *, unsigned int );

/*
			 			  
   Functions for sampling of start values
   --------------------------------------

   TRXParticleSampleSrc    ( particle, src, rnd, flag )
                           x to test
                           o calls 

*/

struct SCMDERR TRXParticleSampleSrc( struct STRXPARTICLE *, struct STRXSRC *, struct STRXRND *, unsigned int );

/*

   Functions for sampling of cross sections
   ----------------------------------------

   TRXParticleSigTot       ( particle, xs, uiFlag )
			                        o calculate total cross section
                         1 . to test
                           o => trxparticle.c
                           o scaling for elastic differently?
			                        o calls

   TRXParticleSampleA      ( particlein, particleout[2], xs, rnd, flag )
                           sample annihilation
                        10 o todo
                           o calls:                          
                             o TRXRNDVRNSD
                             o evtl: 1.5-dim inverse interpolation
                             o TRXParticleChangeDir() to change particle direction

   TRXParticleSampleB      ( particlein, particleout[2], xs, rnd, flag )
                           sample bremsstrahlung 
                         8 o todo
                           o calls:                          
                             o TRXRNDVRNSD
                             o evtl: 1.5-dim inverse interpolation
                             o TRXParticleChangeDir() to change particle direction

   TRXParticleSampleC      ( particlein, particleout[2], xs, rnd, flag )
                             sample compton effect
                         6 o todo
                           ! check coplanarity 
                   		  	   ! re-enable iRC=1 for e- created
                           o calls:                          
                             o TRXRNDVRNSD
                             o evtl: 1.5-dim inverse interpolation
                             o TRXParticleChangeDir() to change particle direction

   TRXParticleSampleF      ( particlein, particleout[2], xs, rnd, flag )
                             sample photo effect
                         7 o todo
                           ! needs info on K-Shell energy, problem for ionic materials
                           o calls:                          
                             o TRXRNDVRNSD
                             o evtl: 1.5-dim inverse interpolation
                             o TRXParticleChangeDir() to change particle direction


   TRXParticleSampleI      ( particlein, particleout[2], xs, rnd, flag )
                           o sample ionization 
                         5 o to test
                     			   o to rework
                             o CW
			                        o proper energy/angular distributions
                           o calls:                          
                             o TRXRNDVRNSD
                             o evtl: 1.5-dim inverse interpolation
                             o TRXParticleChangeDir() to change particle direction
                          
   TRXParticleSampleL      ( particlein, particleout[2], xs, rnd, flag )
                             sample elastic scattering
                         3 . to test
			                        o to complete
                             o photons
			                          o ions
                           o calls:                          
                             o TRXRNDVRNSD
                             o evtl: 1.5-dim inverse interpolation
                             o TRXParticleChangeDir()

   TRXParticleSampleP      ( particlein, particleout[2], xs, rnd, flag )
                           o sample pair production
                         9 o to do
                           o calls:                          
                             o TRXRNDVRNSD
                             o evtl: 1.5-dim inverse interpolation
                             o TRXParticleChangeDir() to change particle direction

   TRXParticleSampleX      ( particlein, particleout[2], xs, rnd, flag )
                           o sample excitation 
                         4 . to test
                           o calls:                          

   TRXParticleSampleTotal  ( particle, xsb, rnd, flag );
                             for given particle type/material
			                          determine total cross section, mean free path,
			                          type of next interaction
                         2 . to test
                           ? handling of out-of-range-high
                           o calls:
                             o TRXXSBaseProjLocate
                             o TRXXSIActTargLocate
                             o TRXParticleSigTot
                             o TRXRNDVRNSD  

   TRXParticleSampleIAct   ( particlein, sPOut[2], xsb, rnd, lFlagXS )
                             sample given interaction/generate secondary particles
                        11 o to test
                           o to rework
                           o returns # of outgoing particles
                           o calls:
                             o TRXXSBaseTargLocate
                             o TRXParticleSampleI()
                             o TRXParticleSampleX()
                             o TRXParticleSampleL()
                             o TRXParticleSampleB()
                             o TRXParticleSampleA()
                             o TRXParticleSampleF()
                             o TRXParticleSampleC()
                             o TRXParticleSampleP()

*/

struct SCMDERR TRXParticleSigTot      ( struct STRXPARTICLE *, struct STRXXS *, unsigned int ); 

struct SCMDERR TRXParticleSampleA     ( struct STRXPARTICLE *, struct STRXPARTICLE [2], struct STRXXS [], struct STRXRND *, unsigned int ); 
struct SCMDERR TRXParticleSampleB     ( struct STRXPARTICLE *, struct STRXPARTICLE [2], struct STRXXS [], struct STRXRND *, unsigned int ); 
struct SCMDERR TRXParticleSampleC     ( struct STRXPARTICLE *, struct STRXPARTICLE [2], struct STRXXS [], struct STRXRND *, unsigned int ); 
struct SCMDERR TRXParticleSampleF     ( struct STRXPARTICLE *, struct STRXPARTICLE [2], struct STRXXS [], struct STRXRND *, unsigned int ); 
struct SCMDERR TRXParticleSampleI     ( struct STRXPARTICLE *, struct STRXPARTICLE [2], struct STRXXS [], struct STRXRND *, unsigned int ); 
struct SCMDERR TRXParticleSampleL     ( struct STRXPARTICLE *, struct STRXPARTICLE [2], struct STRXXS [], struct STRXRND *, unsigned int ); 
struct SCMDERR TRXParticleSampleP     ( struct STRXPARTICLE *, struct STRXPARTICLE [2], struct STRXXS [], struct STRXRND *, unsigned int ); 
struct SCMDERR TRXParticleSampleX     ( struct STRXPARTICLE *, struct STRXPARTICLE [2], struct STRXXS [], struct STRXRND *, unsigned int ); 
struct SCMDERR TRXParticleSampleTotal ( struct STRXPARTICLE *, struct STRXXSBASE *, struct STRXRND *, unsigned int ); 
struct SCMDERR TRXParticleSampleIAct  ( struct STRXPARTICLE *, struct STRXPARTICLE [2], struct STRXXSBASE *, struct STRXRND *, unsigned int );


                                      /* error handling */
int TRXParticleError( struct SCMDERR *, unsigned int );

enum { 
   ERRTRXPARTICLE_SYS=1,
   ERRTRXPARTICLE_BADADDR,
   ERRTRXPARTICLE_BADCOS,
   ERRTRXPARTICLE_BADDIR,
   ERRTRXPARTICLE_NOSUCHSURFACE,
   ERRTRXPARTICLE_NOSUCHVOL,
   ERRTRXPARTICLE_ISECTFAILED,
   ERRTRXPARTICLE_LOCATEFAILED,
   ERRTRXPARTICLE_EOUTOFRANGELO,
   ERRTRXPARTICLE_EOUTOFRANGEHI,
   ERRTRXPARTICLE_BADPROJ,
   ERRTRXPARTICLE_NOMATERIAL,
   ERRTRXPARTICLE_NOSUCHIA,
   ERRTRXPARTICLE_NOTANNIH,
   ERRTRXPARTICLE_NOTBREMS,
   ERRTRXPARTICLE_NOTCOMPTON,
   ERRTRXPARTICLE_NOTPHOTO,
   ERRTRXPARTICLE_NOTIONIZATION,
   ERRTRXPARTICLE_NOTELASTIC,
   ERRTRXPARTICLE_NOTPAIR,
   ERRTRXPARTICLE_NOTEXCITATION,
   ERRTRXPARTICLE_PROJXSMISMATCH,
   ERRTRXPARTICLE_PROJDUPFAILED,
   ERRTRXPARTICLE_SRCPROJFAILED,
   ERRTRXPARTICLE_COMPONENTEXCEEDED,
   ERRTRXPARTICLE_XSINVALID,
   ERRTRXPARTICLE_NOSUCHXSPROJ,
};

#endif
