#ifndef TRXPROJ_H
/** \file
   Various definitions for projectiles.
*/
/*
   Copyright  1992-2015 M.Kraemer, GSI, D-64291 Darmstadt

   $Log: trxproj.h,v $
   Revision 1.4  2015/01/05 13:35:20  kraemer
   streamline, prepare for nuclear

   Revision 1.3  2009/05/29 12:10:22  kraemer
   external projectile tables

   Revision 1.2  2002/11/25 16:33:55  kraemer
   various name <=> macros

   Revision 1.1  2001/01/09 14:54:36  kraemer
   *** empty log message ***


*/
#define TRXPROJ_H

#include "utypes.h"
#include "cmderr.h"

                                      /* masks */
#define TRXPROJ_MASKGLOBAL 0xFF000000 /* global (e.g. debug) */
#define TRXPROJ_MASKFUNC   0x00FF0000 /* function specific flags */
#define TRXPROJ_MASKQUAL   0x0000FFFF /* for additional qualifiers */
                                      /* global flags */
#define TRXPROJ_DEBUG  0x80000000     /* general check flag */
#define TRXPROJ_QUIET  0x10000000     /* suppress error messages */
#define TRXPROJ_NOMSG  TRXPROJ_QUIET
#define TRXPROJ_DUPASREF 0x04000000   /* duplicate as a reference copy, i.e. no allocation */

#define TRXPROJ_R0 1.36               /* default nuclear radius (fm) (Sihver) */ 

                                      /** \struct STRXXSTAG
                                          \brief Aux. structure to correlate tag string <=> ID.
                                      */ 
struct STRXPROJTAGID { char *pcTag; int iID; };

                                      /** \struct STRXZA
                                          \brief Charge & Mass numbers.
                                      */ 
struct STRXZA {
   double dMr;                        /**< mass relative (to AMU/electron mass) */
   double dAb;                        /**< isotopic abundancy */
   double dRMSc, dRMSi;               /**< RMS charge, interaction */
   int iZ;                            /**< charge number */
   int iA;                            /**< mass number (# of nucleons/leptons) */
};
#define TRXZA_A dMr
/*
#define TRXZA_A iA
*/
#define TRXZAMATCH(psZA1,psZA2,iRC)\
   (iRC) = 0;\
   if ( (psZA1)->iA != (psZA2)->iA ) {\
      (iRC) = ERRTRXPROJ_MISMATCHMASS;\
   }\
   if ( (psZA1)->iZ != (psZA2)->iZ ) {\
      (iRC) = ERRTRXPROJ_MISMATCHCHARGE;\
   }


/* 
   Z,A handling functions
   ----------------------

   TRXZAInspect    ( za, flag )
                   o inspect Z,A
                 1 x to test
                   o calls:

   TRXZAMATCH      ( za1, za2, rc )
   TRXZAMatch      ( za1, za2, flag )  
                   o Returns 0 if Z,A's match
                 2 x to test
                   ? should rel.mass and abundancy be tested too ? 
                   o calls:
                     TRXZAMATCH
                   
*/
struct SCMDERR TRXZAMatch  ( struct STRXZA *, struct STRXZA *, unsigned int );
struct SCMDERR TRXZAInspect( struct STRXZA *, unsigned int );

struct STRXZA  _TRXZATest( unsigned int );
                                      /** \struct STRXPROJ
                                          \brief structure to describe a projectile.
                                      */ 
struct STRXPROJ {
   unsigned int uiFlag;               /**< flag bits */
#define TRXPROJ_ISVALID 0x00000001    /* flag valid content */ 
#define TRXPROJ_REFCOPY 0x00000002    /* reference only, do not free() */ 
   unsigned int uiType;               /**< particle type */
   size_t ulSpace;                    /**< space occupied (bytes) */
   char *pcName;                      /**< projectile name (stem) */
   char *pcFullName;                  /**< projectile name (full) */
   int nZA;                           /**< # of isotopes */ 
   struct STRXZA *psZA;               /**< [nZA] Z,A for isotopes */
   double dCharge;                    /**< explicit charge */
   double dM0;                        /**< rest mass [keV] */
   struct STRXZA sZA;                 /**< main Z,A */
};
                                      /* structure initialize, some compilers like it that way */
#define TRXPROJ_INITNULL 0,0,0,NULL,NULL,0,NULL,0.0,0.0,{0.0,0.0,0,0}

enum {
   TRXPROJ_NONE=0, 
   TRXPROJ_N=1, TRXPROJ_X, TRXPROJ_ELECTRON, TRXPROJ_POSITRON, TRXPROJ_ION,
   TRXPROJ_MAX,
   TRXPROJ_ALL=TRXPROJ_MAX
};

enum {
   TRXPROJ_BITN = ( 1L << TRXPROJ_N ), 
   TRXPROJ_BITX = ( 1L << TRXPROJ_X ), 
   TRXPROJ_BITELECTRON  = ( 1L << TRXPROJ_ELECTRON ), 
   TRXPROJ_BITPOSITRON  = ( 1L << TRXPROJ_POSITRON ), 
   TRXPROJ_BITION       = ( 1L << TRXPROJ_ION ), 
};

#define TRXPROJ_NNAME        "n"
#define TRXPROJ_XNAME        "x"
#define TRXPROJ_ELECTRONNAME "e-"
#define TRXPROJ_POSITRONNAME "e+"
#define TRXPROJ_IONNAME      "ion"

#define TRXPROJFILEVERSION 20090415

/* 
   Projectile handling functions
   -----------------------------

   TRXProjDup          ( pto, pfr, flag )
                       o Duplicate a projectile
                     6 x to test
                       o calls
                         o TRXProjInit
                         o TRXProjFree
                         
   TRXProjFree         ( p, flag )
                       o Release a projectile's resource
                     4 x to test
                       o calls
                         o TRXProjInit
                         
   TRXProjFromStr      ( p, str, flag )
   TRXProjFromStrn     ( p, str, nstr, flag )
                       o fill projectile structure from string spec.s 
                     7 x to test
                       o calls:
                         o TRXProjFree
                         o TRXProjInit

   TRXProjInit         ( p, flag )
                       o Initialize a projectile
                     3 x to test
                       o calls
                         o TRXProjSpace
                       
   TRXProjInspect      ( p, flag )
                       o Inspect a projectile
                     1 x to test
                       o calls
                         o TRXZAInspect

   TRXProjList         ( p, flag )
                       o List a projectile
                    13 x to test
                       o to complete
                         o <I> etc
                         o space, charge, restmass
                         o flags for multiple calls from BaseList()
                       o calls
                       
   TRXProjMatch        ( proj1, proj2, flag )
                       o returns 0 if projectiles match
                     8 x to test
                       ? really test charge and restmass 
                       o calls:
                         o TRXZAMATCH

   TRXProjNameFromType ( ptype )
                       o returns pointer to generic projectile name
                     x o to test
                       o calls
                        
   TRXProjRead         ( p, filename, flag )
                       o Read projectile data from file
                    11 x to test
                       o calls:
                         o TRXProjFree
                         o TRXProjInit
                       
   TRXProjSpace        ( p, flag )
                       o Count a projectile's occupied space
                     2 x to test
                       o calls

   TRXProjTypeFromName ( str, len )
                       o returns projectile type from generic projectile name
                    10 x to test
                       o calls
                        
   TRXProjValidate     ( p, flag ) 
                       o check validity
                     5 x to test
                       o calls:
     
   TRXProjWrite        ( p, path, flag )
                       o Write projectile data to file
                    12 x to test
                       o calls


*/
                                      /* function prototypes */
struct SCMDERR TRXProjDup     ( struct STRXPROJ *, struct STRXPROJ *, unsigned int );
struct SCMDERR TRXProjExportCHeader( struct STRXPROJ *, const char *, unsigned int );
struct SCMDERR TRXProjFree    ( struct STRXPROJ *, unsigned int );
struct SCMDERR TRXProjFromStr ( struct STRXPROJ *, const char *, unsigned int );
struct SCMDERR TRXProjFromStrn( struct STRXPROJ *, const char *, size_t, unsigned int );
struct SCMDERR TRXProjInit    ( struct STRXPROJ *, unsigned int );
struct SCMDERR TRXProjInspect ( struct STRXPROJ *, unsigned int );
struct SCMDERR TRXProjList    ( struct STRXPROJ *, unsigned int );
struct SCMDERR TRXProjMatch   ( struct STRXPROJ *, struct STRXPROJ *, unsigned int );
struct SCMDERR TRXProjRead    ( struct STRXPROJ *, const char *, unsigned int );
struct SCMDERR TRXProjSpace   ( struct STRXPROJ *, unsigned int );
struct SCMDERR TRXProjValidate( struct STRXPROJ *, unsigned int );
struct SCMDERR TRXProjWrite   ( struct STRXPROJ *, const char *, unsigned int );

const char *TRXProjNameFromType( unsigned int );
unsigned int TRXProjTypeFromName( const char *, size_t );

struct STRXPROJ _TRXProjTest( int, unsigned int );

                                 
int TRXProjError( struct SCMDERR *, unsigned int );



                                      /** \struct STRXPROJBASE
                                          \brief structure to describe a projectile base.
                                      */ 
struct STRXPROJBASE {
   unsigned int uiFlag;               /**< flag bits */
#define TRXPROJBASE_ISVALID 0x00000001 /* flag valid content */ 
   size_t ulSpace;                    /**< space occupied (bytes) */
   int nP, mP;                        /**< current/max # of projectiles in base */
#define TRXPROJBASE_PROJINC 128       /* mP increment */   
   struct STRXPROJ *psP;              /**< [nP/mP] projectiles */
};
                                      /* projectile slot increment macro */
#define TRXPROJBASEINCREMENT(psPB,sCE,gError) \
    if ( (psPB)->nP >= (psPB)->mP ) {\
       void *pv = NULL;\
       pv = realloc( (psPB)->psP, (ssize_t) ((psPB)->mP+TRXPROJBASE_PROJINC) * sizeof(*((psPB)->psP)) );\
       if ( pv == NULL ) {\
          (sCE).iRC = ERRTRXPROJ_SYS; goto gError;\
       }\
       (psPB)->psP = pv;\
       memset( (psPB)->psP + (psPB)->mP, 0, TRXPROJBASE_PROJINC * sizeof(*((psPB)->psP)));\
       (psPB)->mP += TRXPROJBASE_PROJINC;\
    }                                

/*
   Functions to handle projectile base
   -----------------------------------

   TRXProjBaseDelete ( pb, names, flag )
                     o delete projectiles from base
                  11 x to test
                     o calls
                       o TRXProjFree
                       
   TRXProjBaseFree   ( pb, flag )
                     o release projectile base's resources
                   4 x to test
                     o calls
                       o TRXProjBaseInit
                       o TRXProjFree
                       
   TRXProjBaseInit   ( pb, flag )
                     o initialize projectile base
                   3 x to test
                     o calls
                       o TRXProjBaseSpace
                       
   TRXProjBaseInspect( pb, names, flag )    
                     o inspect projectiles in base
                   1 x to test
                     o calls
                       o TRXProjInspect
                       
   TRXProjBaseList   ( pb, names, flag )    
                     o list projectiles in base
                   9 x to test
                     o calls
                       o TRXProjList
                       
   TRXProjBaseLocateByName( pb, name, flag )
                   x o to test   
                     ? allow full name (to be decomposed by FromStr())
   TRXProjBaseLocate ( pb, p, flag )    
                     o locate a projectile in base
                   6 ? to do
                     o calls
                       o 
                       
   TRXProjBaseRead   ( pb, files, flag )
                     o read projectile files
                   7 o to test
                     o calls
                       o TRXProjFree
                       o TRXProjInit
                       o TRXProjRead
                       o TRXProjBaseLocate
                       
   TRXProjBaseSpace  ( pb, flag )
                     o count space occupied by projectile base
                   2 x to test
                     o calls
                       o TRXProjSpace
                       
                       
   TRXProjBaseValidate( pb, flag )
                     o validate a projectile base
                   5 z to test
                     o calls
                       o TRXProjValidate
                       
   TRXProjBaseWrite  ( pb, prefix, names, flag )    
                     o write projectiles from base
                   8 x to test
                     o calls
                       o TRXProjWrite
                       
   TRXProjCompleteFromBase( p, pb, flag )
                     o locate proj in base, fill entry,
                       most abundant nuclide, weighted mass ?
                  10 x to test
                     o calls  
                       o TRXProjValidate

       
*/
struct SCMDERR TRXProjBaseDelete  ( struct STRXPROJBASE *, const char *, unsigned int );
struct SCMDERR TRXProjBaseExportCHeader( struct STRXPROJBASE *, const char *, const char *, unsigned int );
struct SCMDERR TRXProjBaseFree    ( struct STRXPROJBASE *, unsigned int );
struct SCMDERR TRXProjBaseInit    ( struct STRXPROJBASE *, unsigned int );
struct SCMDERR TRXProjBaseInspect ( struct STRXPROJBASE *, const char *, unsigned int );
struct SCMDERR TRXProjBaseList    ( struct STRXPROJBASE *, const char *, unsigned int );
struct STRXPROJ *TRXProjBaseLocate( struct STRXPROJBASE *, struct STRXPROJ *, unsigned int );
struct STRXPROJ *TRXProjBaseLocateByName ( struct STRXPROJBASE *, const char *, unsigned int );
struct SCMDERR TRXProjBaseRead    ( struct STRXPROJBASE *, const char *, unsigned int );
struct SCMDERR TRXProjBaseSpace   ( struct STRXPROJBASE *, unsigned int );
struct SCMDERR TRXProjBaseValidate( struct STRXPROJBASE *, unsigned int );
struct SCMDERR TRXProjBaseWrite   ( struct STRXPROJBASE *, const char *, const char *, unsigned int );

struct SCMDERR TRXProjCompleteFromBase( struct STRXPROJ *, struct STRXPROJBASE *, unsigned int );
#define TRXPROJCOMPLETEFROMBASE_AVGWEIGHT    0x00010000 /* fill w/ average atomic weight */
#define TRXPROJCOMPLETEFROMBASE_MOSTABUNDANT 0x00020000 /* fill w/ most abundant */

struct STRXPROJBASE _TRXProjBaseTest( unsigned int );

                                      /* command processors */
int TRXCmdProjTest( const char *, int32_t * );

int TRXCmdProj    ( const char *, const char *, const char *, const char *, 
                    int32_t *, int32_t *, int32_t *, int32_t *, int32_t *, int32_t * );
/*
   --- error codes
*/
enum { ERRTRXPROJ_SYS=1,
       ERRTRXPROJ_BADADDR,
       ERRTRXPROJ_BADELEMENT,
       ERRTRXPROJ_BADCHARGE,
       ERRTRXPROJ_MISMATCHMASS,
       ERRTRXPROJ_MISMATCHCHARGE,
       ERRTRXPROJ_MISMATCHPARTICLE,
       ERRTRXPROJ_MISMATCHNAME,
       ERRTRXPROJ_BADZ,
       ERRTRXPROJ_BADA,
       ERRTRXPROJ_BADNAME,
       ERRTRXPROJ_BADZA,
       ERRTRXPROJ_NOSUCHPROJ,
       ERRTRXPROJ_MISMATCHRESTMASS,
       ERRTRXPROJ_BADITEMLIST,
       ERRTRXPROJ_BADFILELIST,
       ERRTRXPROJ_MULTIPLENAME,
       ERRTRXPROJ_BADMASS,
       ERRTRXPROJ_BADABUNDANCY,
       ERRTRXPROJ_READUNKNOWNTAG,
       ERRTRXPROJ_READUNKNOWNTAGDATA,
       ERRTRXPROJ_BADFILETYPE,
       ERRTRXPROJ_BADFILEVERS,
       ERRTRXPROJ_READNOPROJDATA,
       ERRTRXPROJ_SYNTAX,
   ERRTRXPROJ_BASEEMPTY,       
   ERRTRXPROJ_INVALID,
   ERRTRXPROJBASE_INVALID,
};

#endif
